package rangesearch;

import rangesearch.api.*;
import rangesearch.ref.*;

import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.awt.geom.*;
import javax.swing.event.EventListenerList;
import jdsl.core.api.*;
import jdsl.core.ref.*;
import jdsl.geomobj.api.*;
import jdsl.geomobj.ref.*;

/** Manages the display of and user interaction with a RangeModel (a set of
 * points plus a range). */

public class RangeView extends JPanel
    implements MouseListener,
               MouseMotionListener,
               ModelChangedListener {
    protected int m_toolMode = TOOL_NONE;
    protected EventListenerList m_listenerList = new EventListenerList();
    protected RangeModel m_model;
    protected int m_selPoint;
    protected boolean m_draggingRange;
    protected Sequence m_resultPoints;
    protected jdsl.geomobj.api.Point2D m_rangePoint1;

    public static final int TOOL_NONE    = 0;
    public static final int TOOL_POINT   = 1;
    public static final int TOOL_RANGE   = 2;

    private static final int BIG_POINT_WIDTH = 5;

    private static final int SNAP_DIST       = 10;

    public RangeView() {
        super();
        setPreferredSize(new Dimension(400, 400));

        m_model = null;
        m_resultPoints = null;
        m_selPoint = -1;
        m_draggingRange = false;

        setModel(new RangeModel());

        addMouseListener(this);
        addMouseMotionListener(this);
    }

    /** Sets the view's RangeModel. */
    public void setModel(RangeModel model) {
        if(m_model != null) {
            m_model.removeModelChangedListener(this);
        }
        m_model = model;
        if(m_model != null) {
            m_model.addModelChangedListener(this);
        }
        m_resultPoints = null;
        m_selPoint = -1;
        m_draggingRange = false;
        repaint();
    }

    /** Returns the view's current RangeModel. */
    public RangeModel getModel() {
        return m_model;
    }

    /** Sets the collection of result points. */
    public void setResultPoints(Sequence resultPoints) {
        m_resultPoints = resultPoints;
        repaint();
    }

    /** Returns the collection of result points. */
    public Sequence getResultPoints() {
        return m_resultPoints;
    }

    /** Adds a listener for the ToolChangedEvent. */
    public void addToolChangedListener(ToolChangedListener l) {
        m_listenerList.add(ToolChangedListener.class, l);
    }

    /** Removes a listener for the ToolChangedEvent. */
    public void removeToolChangedListener(ToolChangedListener l) {
        m_listenerList.remove(ToolChangedListener.class, l);
    }

    /** Notifies all listeners that the current tool has changed. */
    protected void fireToolChanged() {
        // Guaranteed to return a non-null array
        Object[] listeners = m_listenerList.getListenerList();
        // Process the listeners last to first, notifying
        // those that are interested in this event
        for(int i = listeners.length - 2; i >= 0; i -= 2) {
            if(listeners[i] == ToolChangedListener.class) {
                // Lazily create the event:
                ToolChangedEvent e = new ToolChangedEvent(this, m_toolMode);
                ((ToolChangedListener)listeners[i+1]).toolChanged(e);
            }
        }
    }

    /** Sets the current tool identifier. */
    public void setToolMode(int toolMode) {
        m_toolMode = toolMode;
        if(m_toolMode != TOOL_POINT) {
            m_selPoint = -1;
            repaint();
        }
        fireToolChanged();
    }

    /** Returns the current tool identifier. */
    public int getToolMode() {
        return m_toolMode;
    }

    /** Displays the graph. */
    protected void paintComponent(Graphics g) {
        Graphics2D g2 = (Graphics2D) g;
        g2.setPaint(new Color(192, 192, 192));
    	g2.fillRect(0, 0, getWidth(), getHeight());

        if(m_model == null) {
            return;
        }

        g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
                RenderingHints.VALUE_ANTIALIAS_ON);

        drawPoints(g2);
        drawResultPoints(g2);
        drawRange(g2);
        if(m_toolMode == TOOL_POINT) {
            drawSelPoint(g2);
        }
    }

    /** Draws the points in the model. */
    protected void drawPoints(Graphics2D g2) {
        if(m_model == null) return;

        int pointWidth = BIG_POINT_WIDTH;
        g2.setPaint(Color.black);

        for(int i = 0; i < m_model.getNumPoints(); i ++) {
            // Skip selected point (we draw it later).
            if(i == m_selPoint) continue;

            jdsl.geomobj.api.Point2D p = m_model.getPoint(i);
            g2.fillOval((int) p.x() - pointWidth / 2,
                        (int) p.y() - pointWidth / 2,
                        pointWidth,
                        pointWidth);
        }
    }

    /** Draws the view's result points. */
    protected void drawResultPoints(Graphics2D g2) {
        if(m_model == null) return;
        if(m_resultPoints == null) return;

        int pointWidth = BIG_POINT_WIDTH;
        g2.setPaint(Color.green);

        ObjectIterator oi = m_resultPoints.elements();
        Point2DIterator pi = new Point2DIteratorAdapter(oi);
        while(pi.hasNext()) {
            jdsl.geomobj.api.Point2D p = pi.nextPoint2D();
            g2.fillOval((int) p.x() - pointWidth / 2,
                        (int) p.y() - pointWidth / 2,
                        pointWidth,
                        pointWidth);
        }
    }

    /** Draws the currently selected point. */
    protected void drawSelPoint(Graphics2D g2) {
        if(m_model == null) return;
        if(m_selPoint == -1) return;

        int pointWidth = BIG_POINT_WIDTH;
        g2.setPaint(Color.white);

        jdsl.geomobj.api.Point2D p = m_model.getPoint(m_selPoint);
        g2.fillOval((int) p.x() - pointWidth / 2,
                    (int) p.y() - pointWidth / 2,
                    pointWidth,
                    pointWidth);
    }

    /** Draws the model's range. */
    protected void drawRange(Graphics2D g2) {
        if(m_model == null) return;

        jdsl.geomobj.api.Rectangle2D r = m_model.getRange();
        if(r == null) return;

        Stroke oldStroke = g2.getStroke();
        g2.setStroke(new BasicStroke(1.0f));

        if(m_draggingRange) {
            g2.setPaint(Color.darkGray);
        } else {
            g2.setPaint(Color.black);
        }

        jdsl.geomobj.api.Point2D p1 = r.bottomLeftPoint();
        jdsl.geomobj.api.Point2D p2 = r.topRightPoint();
        int x = (int) p1.x();
        int y = (int) p1.y();
        int w = (int) (p2.x() - p1.x());
        int h = (int) (p2.y() - p1.y());
        g2.drawRect(x, y, w, h);
        g2.setStroke(oldStroke);
    }

    /** Finds the closest point to the given coordinates and sets it to be
     * selected. */
    protected void recalcSelPoint(int x, int y) {
        if(m_model == null) return;
        m_selPoint = -1;
        if(m_toolMode == TOOL_RANGE) return;

        double minDistSq = SNAP_DIST * SNAP_DIST + 1;
        for(int i = 0; i < m_model.getNumPoints(); i ++) {
            jdsl.geomobj.api.Point2D p = m_model.getPoint(i);
            double deltaX = x - p.x();
            double deltaY = y - p.y();
            double distSq = deltaX * deltaX + deltaY * deltaY;
            if(distSq < minDistSq) {
                m_selPoint = i;
                minDistSq = distSq;
            }
        }
    }

    /** Handles model changed events. */
    public void modelChanged(ModelChangedEvent e) {
        m_resultPoints = null;
        repaint();
    }

    /** Empty function for MouseListener. */
    public void mouseClicked(MouseEvent e) {
    }

    /** Empty function for MouseMotionListener. */
    public void mouseEntered(MouseEvent e) {
    }

    /** Empty function for MouseMotionListener. */
    public void mouseExited(MouseEvent e) {
    }

    /** Method for handling mouse presses. */
    public void mousePressed(MouseEvent e) {
        if(m_model == null) return;

        int x = e.getX();
        int y = e.getY();

        recalcSelPoint(x, y);
        if(m_toolMode == TOOL_POINT) {
            if(m_selPoint == -1) {
                m_selPoint = m_model.addPoint(new IntPoint2D(x, y));
            }
        } else if(m_toolMode == TOOL_RANGE) {
            m_model.setRange(null);
            m_rangePoint1 = new IntPoint2D(x, y);
            m_draggingRange = true;
        }
    }

    /** Method for handling mouse releases.  
     */
    public void mouseReleased(MouseEvent e) {
        if(m_model == null) return;

        if(m_toolMode == TOOL_RANGE) {
            m_draggingRange = false;
            repaint();
        }
    }

    /** Method for handling clicked mouse movement.  */
    public void mouseDragged(MouseEvent e) {
        if(m_model == null) return;

        int x = e.getX();
        int y = e.getY();

        if(m_toolMode == TOOL_POINT) {
            m_model.setPoint(m_selPoint, new IntPoint2D(x, y));
        } else if(m_toolMode == TOOL_RANGE) {
            jdsl.geomobj.api.Point2D p2 = new IntPoint2D(x, y);
            try {
                jdsl.geomobj.api.Rectangle2D r;
                r = new IntRectangle2D(m_rangePoint1, p2);
                m_model.setRange(r);
            } catch (InvalidPointException ex) {
                m_model.setRange(null);
            }
        }
    }
    
    /** Method for handling unclicked mouse movement.  Finds the closest point
     * to the cursor and selects it, as appropriate to the current tool mode.
     * */
    public void mouseMoved(MouseEvent e) {
        int oldSelPoint = m_selPoint;
        recalcSelPoint(e.getX(), e.getY());
        if(oldSelPoint != m_selPoint) repaint();
    }
}

